# from .proxy import StorageProxy
#
#
# class ReadKeysCacheRedis(StorageProxy):
#     def __init__(self, storage, redis):
#         super(ReadKeysCacheInMemory, self).__init__(storage)
#         self.redis = redis
#
#     async def create_key(self, storage_key, data=None):
#         data = await self.storage.create_key(storage_key, data)
#         self.t[storage_key.get_url_path()] = ''
#
#         return data
#
#     async def remove_key(self, storage_key):
#         await self.storage.remove_key(storage_key)
#         if storage_key.get_url_path() in self.t:
#             del self.t[storage_key.get_url_path()]
#
#     async def remove_keys(self, storage_keys):
#         await self.storage.remove_keys(storage_keys)
#         for x in storage_keys:
#             self._cache.remove(x.get_url_path())
#
#     async def get_keys(self, prefix, max_keys=200, after=None):
#         resp = None
#         if not after:
#             after = prefix.as_url_path()
#
#         keys = self.t.iterkeys(after)[:max_keys]
#
#         if len(keys) <= max_keys:
#             more = max_keys - len(keys)
#             start_token = keys[-1]
#             resp = self.storage.get_keys(prefix, more, start_token)
#             keys += resp['keys']
#             after = resp.get('after')
#
#         if after:
#             kwargs['after'] = after
#
#         resp = await retry_client(
#             self.get_s3_client().list_objects_v2,
#             **kwargs,
#         )
#
#         response_keys = [x['Key'] for x in resp.get('Contents', [])]
#
#         resp_data = {
#             'keys': response_keys,
#         }
#
#         after = resp.get('after')
#         if after:
#             resp_data['after'] = after
#
#         return resp_data



import pygtrie

from ssshelf.keys import IndexKey

class InMemoryStorage(object):

    def __init__(self):
        self.t = pygtrie.CharTrie()
        self.t.enable_sorting()

    async def create_key(self, storage_key, data=None):
        data = data if data else bytes()

        self.t[storage_key.as_url_path()] = data

        return data

    async def remove_key(self, storage_key):
        del self.t[storage_key.as_url_path()]

    async def remove_keys(self, storage_keys):
        for key in storage_keys:
            del self.t[key.as_url_path()]

    async def get_key(self, storage_key):
        return {
            'data': self.t.get(storage_key.as_url_path()),
            'metadata': {},
        }

    def _build_keys(self, key_iterator, max_keys, after):
        keys = []
        found_token = False

        for x in key_iterator:
            if after and found_token is False:
                if x != after:
                    continue

                found_token = True

                continue

            keys.append(IndexKey.from_url_path(x))

            if len(keys) == max_keys:
                break

        return keys

    async def get_keys(self, prefix, max_keys=200, after=None):
        key_iterator = self.t.iterkeys(prefix.as_url_path())

        try:
            keys = self._build_keys(key_iterator, max_keys, after)
        except KeyError:
            return {
                'keys': []
            }

        try:
            next(key_iterator)
            after = keys[-1].as_url_path()
        except StopIteration:
            after = None

        ret = {
            'keys': keys,
        }

        if after:
            ret['after'] = after

        return ret



import pygtrie

from .proxy import StorageProxy

class ReadKeyCacheInMemory(StorageProxy):
    def __init__(self, storage):
        super(ReadKeyCacheInMemory, self).__init__(storage)
        self._cache = {}

    async def create_key(self, storage_key, data=None):
        data = await self.storage.create_key(storage_key, data)
        self._cache[storage_key.as_url_path()] = {
            'data': data,
        }
        return data

    async def remove_key(self, storage_key):
        await self.storage.remove_key(storage_key)
        del self._cache[storage_key.as_url_path()]

    async def remove_keys(self, storage_keys):
        await self.storage.remove_keys(storage_keys)
        for x in storage_keys:
            if x.as_url_path() in self._cache:
                del self._cache[x.as_url_path()]

    async def get_key(self, storage_key):
        if storage_key.as_url_path() not in self._cache:
            resp = await self.storage.get_key(storage_key)
            self._cache[storage_key.as_url_path()] = resp

        return self._cache[storage_key.as_url_path()]



import attr
import string
from .utils import build_url_path, parse_url_path

ALPHABET = string.ascii_uppercase + string.ascii_lowercase
ALPHABET_REVERSE = dict((c, i) for (i, c) in enumerate(ALPHABET))
BASE = len(ALPHABET)


@attr.s
class IndexKey(object):
    pk = attr.ib()
    index_parts = attr.ib(default=attr.Factory(list))

    def as_url_path(self):
        url_parts = self.index_parts + [self.pk]

        return build_url_path(url_parts)

    @classmethod
    def from_url_path(cls, path):
        url_parts = parse_url_path(path)
        pk = url_parts.pop()

        return cls(pk, url_parts)

@attr.s
class PrefixKey(object):
    index_parts = attr.ib()

    def as_url_path(self):
        url_parts = self.index_parts

        return build_url_path(url_parts)


def encode_int_as_str(n):
    s = []

    while True:
        n, r = divmod(n, BASE)
        s.append(ALPHABET[r])
        if n == 0:
            break

    return ''.join(reversed(s))


def decode_str_as_int(s):
    n = 0

    for c in s:
        n = n * BASE + ALPHABET_REVERSE[c]

    return n


def reverse(s):
    n = []
    for i in s:
        ind = ALPHABET.index(i)
        new_ind = len(ALPHABET) - ind - 1
        n += [ALPHABET[new_ind]]

    return ''.join(n)




class StorageProxy(object):
    def __init__(self, storage):
        self.storage = storage

    async def create_key(self, storage_key, data=None):
        return await self.storage.create_key(storage_key, data)

    async def remove_key(self, storage_key):
        await self.storage.remove_key(storage_key)

    async def remove_keys(self, storage_keys):
        await self.storage.remove_keys(storage_keys)

    async def get_key(self, storage_key):
        return await self.storage.get_key(storage_key)

    async def get_keys(self, *args, **kwargs):
        return await self.storage.get_keys(*args, **kwargs)



import asyncio
import aiobotocore
from retrying import retry

@retry(stop_max_attempt_number=2)
def retry_client(method, *args, **kwargs):
    return method(*args, **kwargs)


class S3Storage(object):

    def __init__(self, bucket, aws_client_kwargs=None, s3_client=None):
        self.aws_client_kwargs = aws_client_kwargs if aws_client_kwargs else {
            'region_name': 'us-east-1',
        }
        self.bucket = bucket
        self.s3_client = s3_client

    def get_s3_client(self):
        if self.s3_client:
            return self.s3_client

        session = aiobotocore.get_session()
        self.s3_client = session.create_client('s3', **self.aws_client_kwargs)

        return self.s3_client

    async def create_key(self, storage_key, data=None):
        data = data if data else bytes()

        await retry_client(
            self.get_s3_client().put_object,
            Bucket=self.bucket,
            Key=storage_key.as_url_path(),
            Body=data
        )

        return data

    async def remove_key(self, storage_key):
        await retry_client(
            self.get_s3_client().delete_object,
            Bucket=self.bucket,
            Key=storage_key.as_url_path(),
        )

    async def remove_keys(self, storage_keys):
        reqs = []
        for key in storage_keys:
            reqs += [self.remove_key(key)]

        resps = [await x for x in reqs]
        # await retry_client(
        #     self.get_s3_client().delete_objects,
        #     Bucket=self.bucket,
        #     Delete={
        #         "Objects": [{'Key': x} for x in storage_keys],
        #         "Quiet": True,
        #     },
        # )

    async def get_key(self, storage_key):
        resp = await retry_client(
            self.get_s3_client().get_object,
            Bucket=self.bucket,
            Key=storage_key.as_url_path(),
        )

        return {
            'data': await resp.get('Body').read(),
            'metadata': resp.get('Metadata', {}),
        }

    async def get_keys(self, prefix, max_keys=200, after=None):
        resp = None

        kwargs = {
            "Bucket": self.bucket,
            "MaxKeys": max_keys,
            "Prefix": prefix.as_url_path(),
        }

        if after:
            kwargs['StartAfter'] = after

        resp = await retry_client(
            self.get_s3_client().list_objects_v2,
            **kwargs,
        )

        response_keys = [x['Key'] for x in resp.get('Contents', [])]

        resp_data = {
            'keys': response_keys,
        }

        after = resp.get('IsTruncated')
        if after:
            resp_data['after'] = resp_data['keys'][-1]

        return resp_data
